import json
from typing import (
    Callable,
    Optional,
    cast,
)

from mypy_boto3_sqs.service_resource import Queue


def read_sqs_json_messages(queue: Queue, predicate: Optional[Callable[[dict], bool]] = None, **kwargs) -> list[dict]:
    defaults = {"MaxNumberOfMessages": 10, "WaitTimeSeconds": 0}
    if kwargs:
        defaults.update(kwargs)

    messages = queue.receive_messages(**defaults)  # type: ignore[arg-type]

    if not messages:
        return []

    parsed = cast(list[dict], [json.loads(message.body) for message in messages])

    if not predicate:
        return parsed

    return [msg for msg in parsed if predicate(msg)]



import json
from collections.abc import Generator
from typing import Any

import petname  # type: ignore[import]
import pytest
from mypy_boto3_dynamodb.service_resource import Table
from mypy_boto3_s3.service_resource import Bucket
from mypy_boto3_s3.type_defs import CreateBucketConfigurationTypeDef
from mypy_boto3_sqs.service_resource import Queue

from nhs_aws_helpers import (
    ddb_table,
    dynamodb,
    events_client,
    s3_client,
    s3_delete_all_versions,
    s3_resource,
    sqs_resource,
)

__all__ = [
    "clone_schema",
    "temp_dynamodb_table",
    "temp_event_bus_fixture",
    "temp_fifo_queue_fixture",
    "temp_queue_fixture",
    "temp_s3_bucket_fixture",
    "temp_s3_bucket_session_fixture",
    "temp_versioned_s3_bucket_fixture",
]


@pytest.fixture(scope="session", name="session_temp_s3_bucket")
def temp_s3_bucket_session_fixture() -> Generator[Bucket, None, None]:
    """
        session temp_s3_bucket ... (use temp_s3_bucket)
    Returns:
        Bucket
    """
    resource = s3_resource()

    bucket_name = f"temp-{petname.generate()}"
    bucket = resource.create_bucket(
        Bucket=bucket_name,
        CreateBucketConfiguration=CreateBucketConfigurationTypeDef(LocationConstraint="eu-west-2"),
    )
    yield bucket

    s3_delete_all_versions(bucket.name, "", dry_run=False)

    bucket.delete()


@pytest.fixture(name="temp_s3_bucket")
def temp_s3_bucket_fixture(session_temp_s3_bucket: Bucket) -> Bucket:
    """
        yields a temporary s3 bucket for use in unit tests

    Returns:
        Bucket: a temporary empty s3 bucket
    """
    bucket = session_temp_s3_bucket

    bucket.objects.all().delete()

    return bucket


@pytest.fixture(name="temp_versioned_s3_bucket")
def temp_versioned_s3_bucket_fixture(session_temp_s3_bucket: Bucket) -> Bucket:
    """
        yields a temporary s3 bucket for use in unit tests

    Returns:
        Bucket: a temporary empty s3 bucket
    """
    bucket = session_temp_s3_bucket

    s3_client().put_bucket_versioning(
        Bucket=bucket.name,
        VersioningConfiguration={
            "MFADelete": "Disabled",
            "Status": "Enabled",
        },
    )

    bucket.objects.all().delete()

    return bucket


@pytest.fixture(name="temp_event_bus")
def temp_event_bus_fixture() -> Generator[tuple[Queue, str], None, None]:
    """
        creates a temporary event bus and a Queue to listen to the event bus
    Returns:
        tuple[Queue, str]: the listening queue and event bus name
    """
    events = events_client()
    sqs = sqs_resource()

    queue_name = f"temp-{petname.Generate(words=2, separator='-')}"
    queue = sqs.create_queue(QueueName=queue_name)

    bus_name = f"temp-{petname.generate()}"
    events.create_event_bus(Name=bus_name)

    rule_name = f"temp-{petname.generate()}"
    events.put_rule(Name=rule_name, EventPattern=json.dumps({"account": ["000000000000"]}), EventBusName=bus_name)

    queue_arn = f"arn:aws:sqs:eu-west-2:000000000000:{queue_name}"
    target_id = f"temp-{petname.generate()}"
    events.put_targets(
        Targets=[{"Id": target_id, "Arn": queue_arn}],
        EventBusName=bus_name,
        Rule=rule_name,
    )

    yield queue, bus_name

    events.remove_targets(Rule=rule_name, EventBusName=bus_name, Ids=[target_id], Force=True)

    events.delete_rule(Name=rule_name, EventBusName=bus_name, Force=True)
    queue.delete()
    events.delete_event_bus(Name=bus_name)


@pytest.fixture(name="temp_queue")
def temp_queue_fixture() -> Generator[Queue, None, None]:
    """
        yields a temporary SQS queue for testing
    Returns:
        Queue
    """
    sqs = sqs_resource()

    queue_name = f"local-{petname.Generate(words=2, separator='-')}"
    queue = sqs.create_queue(QueueName=queue_name, Attributes={"VisibilityTimeout": "2"})

    yield queue

    queue.delete()


@pytest.fixture(name="temp_fifo_queue")
def temp_fifo_queue_fixture() -> Generator[Queue, None, None]:
    """
        yields a temporary Fifo SQS queue for testing
    Returns:
        Queue
    """
    sqs = sqs_resource()

    queue_name = f"local-{petname.Generate(words=2, separator='-')}.fifo"
    queue = sqs.create_queue(QueueName=queue_name, Attributes={"FifoQueue": "true", "VisibilityTimeout": "2"})

    yield queue

    queue.delete()


def clone_schema(table, on_demand_billing_mode: bool = True, provisioned_capacity: int = 250):
    key_schema = table.key_schema

    attributes = table.attribute_definitions

    indexes = table.global_secondary_indexes

    if indexes:
        for index in indexes:
            del index["IndexStatus"]
            del index["IndexSizeBytes"]
            del index["ItemCount"]
            del index["IndexArn"]
            if "ProvisionedThroughput" in index:
                del index["ProvisionedThroughput"]

    clone = {
        "KeySchema": key_schema,
        "AttributeDefinitions": attributes,
    }
    if on_demand_billing_mode:
        provisioned_capacity = 0
    billing: dict[str, Any] = (
        {
            "BillingMode": "PAY_PER_REQUEST",
        }
        if on_demand_billing_mode
        else {
            "ProvisionedThroughput": {
                "ReadCapacityUnits": provisioned_capacity,
                "WriteCapacityUnits": provisioned_capacity,
            }
        }
    )

    clone.update(billing)

    if indexes:
        if not on_demand_billing_mode:
            for index in indexes:
                index.update(billing)
        clone["GlobalSecondaryIndexes"] = indexes

    return clone


def temp_dynamodb_table(
    source_table_name: str, on_demand_billing_mode: bool = True, provisioned_capacity: int = 500
) -> Generator[Table, None, None]:
    """
    Create a table that copies the schema of <source_table> but uses a random name, can be used throughout
    a test and is deleted at the end.
    """
    ddb = dynamodb()

    table_name = f"pytest-{petname.Generate(words=4, separator='_')}"

    source_table = ddb_table(source_table_name)

    cloned = clone_schema(source_table, on_demand_billing_mode, provisioned_capacity)

    table = ddb.create_table(TableName=table_name, **cloned)

    yield table

    table.delete()



import asyncio
import contextlib
import gzip
import inspect
import io
import os
import random
import re
import time
from collections.abc import AsyncGenerator, Generator, Iterable, Sequence
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial, reduce, wraps
from typing import (
    IO,
    Any,
    Callable,
    Literal,
    Optional,
    TypeVar,
    Union,
    cast,
)
from urllib import parse
from uuid import uuid4

import boto3
import botocore.credentials
import botocore.session
from boto3.dynamodb.types import TypeDeserializer
from boto3.s3.transfer import TransferConfig
from boto3.session import Session
from botocore.config import Config
from botocore.exceptions import ClientError, UnknownServiceError
from botocore.response import StreamingBody
from botocore.retries import quota
from botocore.retries.base import BaseRetryBackoff
from botocore.retries.standard import (
    DEFAULT_MAX_ATTEMPTS,
    RetryContext,
    RetryEventAdapter,
    RetryHandler,
    RetryPolicy,
    RetryQuotaChecker,
    StandardRetryConditions,
)
from mypy_boto3_athena import AthenaClient
from mypy_boto3_dynamodb.client import DynamoDBClient
from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource, Table
from mypy_boto3_dynamodb.type_defs import KeysAndAttributesTypeDef
from mypy_boto3_events import EventBridgeClient
from mypy_boto3_firehose import FirehoseClient
from mypy_boto3_kms.client import KMSClient
from mypy_boto3_lambda.client import LambdaClient
from mypy_boto3_logs.client import CloudWatchLogsClient
from mypy_boto3_s3.client import S3Client
from mypy_boto3_s3.service_resource import Bucket, Object, ObjectSummary, ObjectVersion, S3ServiceResource
from mypy_boto3_s3.type_defs import (
    CompletedPartTypeDef,
    DeleteMarkerEntryTypeDef,
    DeleteTypeDef,
    GetObjectOutputTypeDef,
    ObjectVersionTypeDef,
)
from mypy_boto3_secretsmanager.client import SecretsManagerClient
from mypy_boto3_ses import SESClient
from mypy_boto3_sns.client import SNSClient
from mypy_boto3_sqs import SQSServiceResource
from mypy_boto3_sqs.client import SQSClient
from mypy_boto3_ssm.client import SSMClient
from mypy_boto3_stepfunctions import SFNClient

from nhs_aws_helpers.common import run_in_executor
from nhs_aws_helpers.s3_object_writer import S3ObjectWriter

s3re = re.compile(r"^(s3[an]?)://([^/]+)/(.+)$", re.IGNORECASE)


def s3_build_uri(bucket: str, key: str) -> str:
    return f"s3://{bucket}/{key}"


def s3_split_path(s3uri: str) -> tuple[str, str, str]:
    match = s3re.match(s3uri)
    if not match:
        raise ValueError(f"Not a s3 uri: {s3uri}")
    scheme, bucket, key = match.groups()
    return scheme, bucket, key


_default_configs: dict[str, Config] = {}

_pre_configure: Optional[Callable[[str, str, Optional[Config]], Optional[Config]]] = None


def register_config_default(service: str, config: Config):
    """
        register default config, that will be used or merged for a given aws service
    Args:
        service: single service e.g s3, dynamodb .. or * for all services
        config: boto config
    """
    _default_configs[service] = config


_post_create_client: Optional[Callable[[str, str, object], Any]] = None


def post_create_client(post_create: Callable[[str, str, object], Any]):
    """
        hook to allow configuration of a created service after created
    Args:
        post_create: callable, takes args  aws service, resource/client, created client
    """
    global _post_create_client
    _post_create_client = post_create


def _aws(
    aws_module: str,
    boto_method: Literal["client", "resource"],
    session: Optional[Session],
    config: Optional[Config],
):
    """
    Core function to construct AWS client/resource objects.

    `session` should be:
    - None
    - a session initialized to ptl
    - a session initialized to prod
    see sessions.py
    """
    session_or_module = boto3 if session is None else session

    cls = getattr(session_or_module, boto_method)

    default_config = _default_configs.get(aws_module, _default_configs.get("*"))
    if default_config:
        config = config.merge(default_config) if config else default_config

    aws_endpoint_url = os.environ.get("AWS_ENDPOINT_URL")
    aws_region = os.environ.get("AWS_REGION", "eu-west-2")

    kwargs = {"region_name": aws_region, "config": config}
    if aws_endpoint_url is not None:  # For localstack (None = AWS)
        kwargs.update(
            {
                "endpoint_url": aws_endpoint_url,
                "aws_access_key_id": "abc",
                "aws_secret_access_key": "123",
                "verify": False,  # type: ignore[dict-item]
            }
        )
    created = cls(aws_module, **kwargs)

    if _post_create_client:
        _post_create_client(aws_module, boto_method, created)

    return created


def cloudwatchlogs_client(session: Optional[Session] = None, config: Optional[Config] = None) -> CloudWatchLogsClient:
    return _aws("logs", "client", session, config)  # type: ignore[no-any-return]


def athena_client(session: Optional[Session] = None, config: Optional[Config] = None) -> AthenaClient:
    return _aws("athena", "client", session, config)  # type: ignore[no-any-return]


def lambdas(session: Optional[Session] = None, config: Optional[Config] = None) -> LambdaClient:
    return _aws("lambda", "client", session, config)  # type: ignore[no-any-return]


def stepfunctions(session: Optional[Session] = None, config: Optional[Config] = None) -> SFNClient:
    return _aws("stepfunctions", "client", session, config)  # type: ignore[no-any-return]


def s3_resource(session: Optional[Session] = None, config: Optional[Config] = None) -> S3ServiceResource:
    return _aws("s3", "resource", session, config)  # type: ignore[no-any-return]


def s3_client(session: Optional[Session] = None, config: Optional[Config] = None) -> S3Client:
    return _aws("s3", "client", session, config)  # type: ignore[no-any-return]


def register_retry_handler(
    client_or_resource: Union[S3ServiceResource, S3Client],
    on_error: Optional[Callable] = None,
    on_backoff: Optional[Callable[[int, RetryContext], Any]] = None,
):
    if not hasattr(client_or_resource.meta, "events"):
        # it's a resource, so get the client and work with that
        client_or_resource = cast(S3ServiceResource, client_or_resource).meta.client

    client = cast(S3Client, client_or_resource)

    retry_quota = RetryQuotaChecker(quota.RetryQuota())

    max_attempts = client.meta.config.retries.get("total_max_attempts", DEFAULT_MAX_ATTEMPTS)  # type: ignore[attr-defined]

    service_id = client.meta.service_model.service_id
    service_event_name = service_id.hyphenize()

    handler = RetryHandler(
        retry_policy=RetryPolicy(
            retry_checker=StandardRetryConditions(max_attempts=max_attempts),
            retry_backoff=CustomBackoff(on_backoff=on_backoff),
        ),
        retry_event_adapter=RetryEventAdapter(),
        retry_quota=retry_quota,  # type: ignore[arg-type]
    )

    unique_id = f"retry-config-{service_event_name}"

    # Unregister the retry config that was created by default
    client.meta.events.unregister(
        f"needs-retry.{service_event_name}",
        unique_id=unique_id,
    )

    # Re-register with our own handler
    client.meta.events.register(
        f"needs-retry.{service_event_name}",
        handler.needs_retry,  # type: ignore[arg-type]
        unique_id=unique_id,
    )

    def on_response_received(**kwargs):
        if (on_error is not None and kwargs.get("exception")) or kwargs.get("parsed_response", {}).get("Error"):
            assert on_error
            on_error(**kwargs)

    if on_error:
        client.meta.events.register(f"response-received.{service_event_name}", on_response_received)

    return handler


class CustomBackoff(BaseRetryBackoff):
    _BASE = 2
    _MAX_BACKOFF = 10

    def __init__(
        self,
        max_backoff=_MAX_BACKOFF,
        rand=random.random,
        on_backoff: Optional[Callable[[int, RetryContext], Any]] = None,
    ):
        self._base = self._BASE
        self._max_backoff = max_backoff
        self._random = rand
        self._on_backoff = on_backoff

    def delay_amount(self, context: RetryContext):
        # The context.attempt_number is a 1-based value, but we have to calculate the delay based on aa 0-based value.
        # First delay set to 0 to retry immediately - the remainder are jittered by the random.
        backoff = min(
            0 if context.attempt_number == 1 else (self._random() * (self._base ** (context.attempt_number - 1))),
            self._max_backoff,
        )

        if self._on_backoff:
            self._on_backoff(backoff, context)

        return backoff


def dynamodb(session: Optional[Session] = None, config: Optional[Config] = None) -> DynamoDBServiceResource:
    return _aws("dynamodb", "resource", session, config)  # type: ignore[no-any-return]


def dynamodb_client(session: Optional[Session] = None, config: Optional[Config] = None) -> DynamoDBClient:
    return _aws("dynamodb", "client", session, config)  # type: ignore[no-any-return]


def ddb_table(table_name: str, session: Optional[Session] = None, config: Optional[Config] = None) -> Table:
    return dynamodb(session=session, config=config).Table(table_name)


def ses_client(session: Optional[Session] = None, config: Optional[Config] = None) -> SESClient:
    return _aws("ses", "client", session, config)  # type: ignore[no-any-return]


def sqs_client(session: Optional[Session] = None, config: Optional[Config] = None) -> SQSClient:
    return _aws("sqs", "client", session, config)  # type: ignore[no-any-return]


def sqs_resource(session: Optional[Session] = None, config: Optional[Config] = None) -> SQSServiceResource:
    return _aws("sqs", "resource", session, config)  # type: ignore[no-any-return]


def ssm_client(session: Optional[Session] = None, config: Optional[Config] = None) -> SSMClient:
    return _aws("ssm", "client", session, config)  # type: ignore[no-any-return]


def firehose_client(session: Optional[Session] = None, config: Optional[Config] = None) -> FirehoseClient:
    return _aws("firehose", "client", session, config)  # type: ignore[no-any-return]


def kms_client(session: Optional[Session] = None, config: Optional[Config] = None) -> KMSClient:
    return _aws("kms", "client", session, config)  # type: ignore[no-any-return]


def secrets_client(session: Optional[Session] = None, config: Optional[Config] = None) -> SecretsManagerClient:
    return _aws("secretsmanager", "client", session, config)  # type: ignore[no-any-return]


def secret_value(name: str, session: Optional[Session] = None, config: Optional[Config] = None) -> str:
    return secrets_client(session=session, config=config).get_secret_value(SecretId=name)["SecretString"]


def secret_binary_value(name: str, session: Optional[Session] = None, config: Optional[Config] = None) -> bytes:
    return secrets_client(session=session, config=config).get_secret_value(SecretId=name)["SecretBinary"]


def ssm_parameter(
    name: str, decrypt=False, session: Optional[Session] = None, config: Optional[Config] = None
) -> Union[str, list[str]]:
    ssm = ssm_client(session=session, config=config)
    value = cast(Union[str, list[str]], ssm.get_parameter(Name=name, WithDecryption=decrypt)["Parameter"]["Value"])
    return value


def sns_client(session: Optional[Session] = None, config: Optional[Config] = None) -> SNSClient:
    return _aws("sns", "client", session, config)  # type: ignore[no-any-return]


def events_client(session: Optional[Session] = None, config: Optional[Config] = None) -> EventBridgeClient:
    return _aws("events", "client", session, config)  # type: ignore[no-any-return]


def s3_bucket(bucket: str, session: Optional[Session] = None, config: Optional[Config] = None) -> Bucket:
    return s3_resource(session=session, config=config).Bucket(bucket)


def s3_object(
    bucket_or_url: str, key: Optional[str] = None, session: Optional[Session] = None, config: Optional[Config] = None
) -> Object:
    if key is not None:
        bucket = bucket_or_url
    else:
        url_parsed = parse.urlparse(bucket_or_url)
        bucket = url_parsed.netloc
        key = url_parsed.path.lstrip("/")

    return s3_resource(session=session, config=config).Object(bucket, key)


def s3_get_tags(
    bucket: str,
    key: str,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    client: Optional[S3Client] = None,
) -> dict[str, str]:
    client = client or s3_resource(session=session, config=config).meta.client
    result = client.get_object_tagging(Bucket=bucket, Key=key)
    tags = {pair["Key"]: pair["Value"] for pair in result["TagSet"]}
    return tags


def s3_replace_tags(
    bucket: str,
    key: str,
    tags: dict[str, str],
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    client: Optional[S3Client] = None,
):
    client = client or s3_resource(session=session, config=config).meta.client

    client.put_object_tagging(
        Bucket=bucket, Key=key, Tagging={"TagSet": [{"Key": k, "Value": v} for k, v in tags.items()]}
    )


def s3_update_tags(
    bucket: str,
    key: str,
    tags: dict[str, str],
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    client: Optional[S3Client] = None,
) -> dict[str, str]:
    client = client or s3_resource(session=session, config=config).meta.client

    result = client.get_object_tagging(Bucket=bucket, Key=key)
    existing_tags = {pair["Key"]: pair["Value"] for pair in result["TagSet"]}
    existing_tags.update(tags)

    client.put_object_tagging(
        Bucket=bucket, Key=key, Tagging={"TagSet": [{"Key": k, "Value": v} for k, v in existing_tags.items()]}
    )
    return existing_tags


def s3_get_all_keys(
    bucket: str,
    prefix: str,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    client: Optional[S3Client] = None,
) -> list[str]:
    client = client or s3_resource(session=session, config=config).meta.client
    paginator = client.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)
    keys = []
    for page in page_iterator:
        keys.extend([content["Key"] for content in page.get("Contents", [])])

    return keys


def s3_delete_keys(
    keys: Iterable[str], bucket: str, session: Optional[Session] = None, config: Optional[Config] = None
):
    buck = s3_bucket(bucket, session=session, config=config)

    all_keys = list(keys)

    if not all_keys:
        return []

    deleted = []

    batch = all_keys[:999]
    remaining = all_keys[999:]
    while batch:
        buck.delete_objects(Delete=cast(DeleteTypeDef, {"Objects": [{"Key": key} for key in batch]}))
        deleted.extend(batch)
        batch = remaining[:999]
        remaining = remaining[999:]

    return deleted


def s3_delete_versioned_keys(
    keys: Iterable[tuple[str, str]], bucket: str, session: Optional[Session] = None, config: Optional[Config] = None
):
    # delete specific versions, rather than deleting "objects" and adding delete_marker
    buck = s3_bucket(bucket, session=session, config=config)

    all_keys = list(keys)
    deleted = []

    # batch deleting can only handle certain sized batches
    batch = all_keys[:999]

    remaining = all_keys[999:]
    while batch:
        buck.delete_objects(
            Delete=cast(
                DeleteTypeDef, {"Objects": [{"Key": key, "VersionId": version_id} for key, version_id in batch]}
            )
        )
        deleted.extend(batch)
        batch = remaining[:999]
        remaining = remaining[999:]

    return deleted


def s3_delete_all_versions(
    bucket: str,
    prefix: Optional[str] = None,
    predicate=None,
    dry_run: bool = True,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
):
    client = s3_client(session=session, config=config)
    object_response_paginator = client.get_paginator("list_object_versions")

    delete_marker_list = []
    version_list = []

    paginator = (
        object_response_paginator.paginate(Bucket=bucket, Prefix=prefix)
        if prefix
        else object_response_paginator.paginate(Bucket=bucket)
    )

    def filter_items(items: Sequence[Union[DeleteMarkerEntryTypeDef, ObjectVersionTypeDef]]):
        for item in items:
            if predicate and not predicate(item["Key"]):
                continue
            yield {"Key": item["Key"], "VersionId": item["VersionId"]}

    for object_response_itr in paginator:
        delete_marker_list.extend(filter_items(object_response_itr.get("DeleteMarkers") or []))
        version_list.extend(filter_items(object_response_itr.get("Versions") or []))

    if dry_run:
        print("dry run found:")
        print(delete_marker_list)
        print(version_list)
        return

    for i in range(0, len(delete_marker_list), 1000):
        response = client.delete_objects(
            Bucket=bucket,
            Delete={"Objects": delete_marker_list[i : i + 1000], "Quiet": True},  # type: ignore[typeddict-item]
        )
        print(response)

    for i in range(0, len(version_list), 1000):
        response = client.delete_objects(
            Bucket=bucket,
            Delete={"Objects": version_list[i : i + 1000], "Quiet": True},  # type: ignore[typeddict-item]
        )
        print(response)


def s3_ls(
    uri: str,
    recursive: bool = True,
    predicate: Optional[Callable[[str], bool]] = None,
    versioning: bool = False,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
) -> Generator[Union[ObjectSummary, ObjectVersion], None, None]:
    _, bucket, path = s3_split_path(uri)

    yield from s3_list_bucket(
        bucket, path, recursive=recursive, predicate=predicate, versioning=versioning, session=session, config=config
    )


def s3_list_bucket(
    bucket: str,
    prefix: str,
    recursive: bool = True,
    predicate: Optional[Callable[[str], bool]] = None,
    versioning: bool = False,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
) -> Generator[Union[ObjectSummary, ObjectVersion], None, None]:
    """list contents of S3 bucket based on filter criteria and versioning flag

    Args:
        bucket (str): bucket name to list contents of
        prefix (str): prefix to filter on
        recursive (bool): whether to recurse or not
        predicate (Callable[[str], bool]): predicate function to filter results on
        versioning (bool): whether to return objects or versions
        session (Session): optional existing session r
        config (Config): optional botocore config

    Returns:
        Generator[object, None, None]: resulting objects (ObjectSummary) or versions (ObjectVersion)
    """
    buck = s3_bucket(bucket, session=session, config=config)
    bc_objects = buck.object_versions if versioning else buck.objects

    for s3_obj in bc_objects.filter(Prefix=prefix, Delimiter="" if recursive else "/"):
        if predicate and not predicate(s3_obj.key):
            continue
        yield s3_obj


def s3_list_delete_markers(
    bucket: str,
    prefix: str,
    recursive: bool = True,
    predicate: Optional[Callable[[str], bool]] = None,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    client: Optional[S3Client] = None,
) -> Generator[DeleteMarkerEntryTypeDef, None, None]:
    """list contents of S3 bucket based on filter criteria and versioning flag

    Args:
        bucket (str): bucket name to list contents of
        prefix (str): prefix to filter on
        recursive (bool): whether to recurse or not
        predicate (Callable[[str], bool]): predicate function to filter results on
        session (Session): optional existing session r
        config (Config): optional botocore config
        client (S3Client): optional pre created client
    Returns:
        Generator[object, None, None]: resulting objects or versions
    """
    client = client or s3_client(session=session, config=config)
    paginator = client.get_paginator("list_object_versions").paginate(
        Bucket=bucket, Prefix=prefix, Delimiter="" if recursive else "/"
    )
    for page in paginator:
        delete_markers = page.get("DeleteMarkers", [])
        for delete_marker in delete_markers:
            if predicate and not predicate(delete_marker["Key"]):
                continue
            yield delete_marker


def s3_put(bucket: Bucket, key: str, body: Union[bytes, str], encoding: str = "utf-8", **kwargs) -> Object:
    obj = bucket.Object(key)
    encoded = cast(bytes, body)
    if isinstance(body, str):
        encoded = cast(str, body).encode(encoding=encoding)

    obj.put(Body=encoded, **kwargs)
    return obj


def s3_list_prefixes(s3_path: str, session: Optional[Session] = None, config: Optional[Config] = None) -> list[str]:
    _na, bucket, prefix = s3_split_path(s3_path)

    if not prefix.endswith("/"):
        prefix += "/"
    client = s3_client(session=session, config=config)
    result = client.list_objects(Bucket=bucket, Prefix=prefix, Delimiter="/")

    return [o["Prefix"].replace(prefix, "").strip("/") for o in result.get("CommonPrefixes", [])]


def s3_list_folders(bucket_name: str, bucket_prefix: str, page_size: int = 100) -> list[str]:
    paginator = s3_client().get_paginator("list_objects")
    folders = []
    iterator = paginator.paginate(
        Bucket=bucket_name, Prefix=bucket_prefix, Delimiter="/", PaginationConfig={"PageSize": page_size}
    )
    for response_data in iterator:
        prefixes = response_data.get("CommonPrefixes", [])
        for prefix in prefixes:
            prefix_name = prefix["Prefix"]
            if prefix_name.endswith("/"):
                folders.append(prefix_name.rstrip("/"))

    return folders


def s3_uri_get_size(s3_uri: str, session: Optional[Session] = None, config: Optional[Config] = None) -> int:
    """
    Get the size in bytes of the file(s) at/under the provided uri
    """
    _, bucket, key = s3_split_path(s3_uri)
    return s3_get_size(bucket, key, session, config)  # type: ignore[no-any-return]


def s3_get_size(
    bucket_name: str, prefix: str, session: Optional[Session] = None, config: Optional[Config] = None
) -> int:
    """
    Get the size in bytes of the file(s) at/under the provided prefix
    """
    size = sum(
        s3_obj.size for s3_obj in s3_bucket(bucket_name, session=session, config=config).objects.filter(Prefix=prefix)
    )
    return size


def s3_upload_multipart_from_copy(
    obj: Object, parts_keys: Sequence[str], executor_type: type[ThreadPoolExecutor] = ThreadPoolExecutor, **kwargs
):
    multipart_upload = obj.initiate_multipart_upload(**kwargs)

    outstanding_part_keys = set(parts_keys)

    # We expect timeouts and other intermittents (hence multiple attempts) but we don't want to log "failed" actions
    # when that happens. We register ClientErrors as "expected", so they are captured with an "error" status instead.
    # Any non-ClientError exceptions will be logged as "failed"
    def copy_part(part_key: str, _attempt: int):
        index = parts_keys.index(part_key) + 1
        multipart_upload.Part(index).copy_from(  # type: ignore[arg-type]
            CopySource={"Bucket": obj.bucket_name, "Key": part_key}
        )

        return part_key

    with executor_type() as executor:
        attempt = 0
        max_retries = 4

        while outstanding_part_keys:
            attempt += 1
            futures = [executor.submit(copy_part, key, attempt) for key in outstanding_part_keys]
            for future in as_completed(futures):
                with contextlib.suppress(BaseException):
                    # Exception will be logged on copy_part()
                    # This will raise an exception if a future has raised an exception
                    key = future.result()
                    outstanding_part_keys.remove(key)

            if outstanding_part_keys and attempt > max_retries:
                raise OSError(f"Failed to multipart_upload {outstanding_part_keys} after {attempt} attempts")

    return multipart_upload.complete(
        MultipartUpload={
            "Parts": [
                CompletedPartTypeDef(ETag=part.e_tag, PartNumber=part.part_number)  # type: ignore[typeddict-item]
                for part in multipart_upload.parts.all()
            ]
        },
    )


def s3_copy_object(
    source_url: str,
    destination_url: str,
    version_id: Optional[str] = None,
    extra_args: Optional[dict] = None,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
):
    _, src_bucket_name, src_key = s3_split_path(source_url)
    _, dest_bucket_name, dest_key = s3_split_path(destination_url)

    copy_source = {"Bucket": src_bucket_name, "Key": src_key}
    if version_id is not None:
        copy_source["VersionId"] = version_id

    destination_buck = s3_bucket(dest_bucket_name, session=session, config=config)
    destination_buck.copy(copy_source, dest_key, ExtraArgs=extra_args or {})  # type: ignore[arg-type]


def s3_upload_file(
    file_path: str, bucket: str, key: str, session: Optional[Session] = None, config: Optional[Config] = None
):
    """
    Upload a file to a given S3 location
    Args:
        file_path: Path to file to be uploaded
        bucket: Target bucket
        key: Target key
        session: boto3 session
        config: optional botocore config
    """
    client = s3_resource(session=session, config=config).meta.client
    client.upload_file(file_path, bucket, key)


def s3_upload_data(
    bucket: str, key: str, data: str, session: Optional[Session] = None, config: Optional[Config] = None
):
    """
    Upload a file to a given S3 location
    Args:
        bucket: Target bucket
        key: Target key
        data: data to be uploaded
        session: boto3 session
        config: optional botocore config
    """
    s3object = s3_resource(session=session, config=config).Object(bucket, key)
    s3object.put(Body=data.encode())


def s3_download_file(
    file_path: str, bucket: str, key: str, session: Optional[Session] = None, config: Optional[Config] = None
):
    """
    Download a file from a given S3 location
    Args:
        file_path: Path to file to be uploaded
        bucket: Target bucket
        key: Target key
        session: boto3 session
        config: optional botocore config
    """
    client = s3_resource(session=session, config=config).meta.client
    client.download_file(bucket, key, file_path)


def s3_download_fileobj(
    file_obj: Union[IO[Any], StreamingBody],
    bucket: str,
    key: str,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    transfer_config: Optional[TransferConfig] = None,
):
    """
    Download a file from a given S3 location
    Args:
        file_obj: The file object to stream to
        bucket: Target bucket
        key: Target key
        session: boto3 session
        transfer_config: transfer config to specify threading, chunk size, etc. if desired
        config: optional botocore config
    """
    client = s3_resource(session=session, config=config).meta.client
    if transfer_config:
        client.download_fileobj(bucket, key, file_obj, Config=transfer_config)
    else:
        client.download_fileobj(bucket, key, file_obj)


def s3_create_presigned_url(
    bucket: str, key: str, expiration: int = 3600, session: Optional[Session] = None, config: Optional[Config] = None
):
    """
    Creates a presigned url for a given S3 object
    Args:
        bucket: Target bucket
        key: Target key
        expiration: Time in seconds for the presigned URL to remain valid
        session: boto3 session
        config: optional botocore config

    Returns:
        Presigned URL as string
    """
    # why not s3_client helper ?
    client = s3_resource(session=session, config=config).meta.client
    params = {"Bucket": bucket, "Key": key}
    response = client.generate_presigned_url("get_object", Params=params, ExpiresIn=expiration)
    return response


_RETRY_EXCEPTIONS = ("ProvisionedThroughputExceededException", "ThrottlingException", "TransactionConflictException")
_TRANSACTION_CANCELLED = "TransactionCanceledException"
_TRANSACTION_CONFLICT = "TransactionConflict"

FuncT = TypeVar("FuncT", bound=Callable[..., Any])


class DDBRetryBackoff:
    def __init__(self, max_retries: int = 6, backoff: float = 0.01, jitter_ratio: float = 0.1):
        self._max_retries = max_retries
        self._backoff = backoff
        self._jitter_ratio = jitter_ratio
        if jitter_ratio < 0 or jitter_ratio > 0.5:
            raise ValueError(f"jitter ratio should be between 0 and 0.5, actual {jitter_ratio}")

    def _should_retry(self, err: ClientError, retries: int):
        if retries >= self._max_retries:
            return False

        reasons = transaction_cancelled_for_only_reasons(err, _TRANSACTION_CONFLICT)
        if reasons:
            return True

        return err.response["Error"]["Code"] in _RETRY_EXCEPTIONS

    def _sleep_for(self, retries: int):
        backoff = pow(2, retries) * self._backoff
        jitter_ratio = 1 + (random.choice([1, -1]) * self._jitter_ratio)
        return backoff * jitter_ratio

    def __call__(self, func: FuncT) -> FuncT:
        @wraps(func)
        def _wrapper(*args, **kwargs):
            retries = 0
            while True:
                try:
                    result = func(*args, **kwargs)

                    return result

                except ClientError as err:  # noqa: PERF203
                    if not self._should_retry(err, retries):
                        raise

                    retries += 1
                    time.sleep(self._sleep_for(retries))

        @wraps(func)
        async def _async_wrapper(*args, **kwargs):
            retries = 0
            while True:
                try:
                    result = await func(*args, **kwargs)

                    return result

                except ClientError as err:  # noqa: PERF203
                    if not self._should_retry(err, retries):
                        raise

                    retries += 1
                    await asyncio.sleep(self._sleep_for(retries))

        if inspect.iscoroutinefunction(func):
            return cast(FuncT, _async_wrapper)

        return cast(FuncT, _wrapper)


dynamodb_retry_backoff = DDBRetryBackoff


@dynamodb_retry_backoff()
def get_items_batched(
    ddb_table_name: str,
    keys: list[dict[str, Any]],
    client: Optional[DynamoDBClient] = None,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    **kwargs,
):
    client = client or cast(DynamoDBClient, dynamodb(session=session, config=config).meta.client)
    response = client.batch_get_item(
        RequestItems={ddb_table_name: cast(KeysAndAttributesTypeDef, {"Keys": keys})}, **kwargs
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200
    return response


def ddb_get_items(
    ddb_table_name: str,
    keys: list[dict[str, Any]],
    client: Optional[DynamoDBClient] = None,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    **kwargs,
) -> list[dict[str, Any]]:
    client = client or cast(DynamoDBClient, dynamodb(session=session, config=config).meta.client)
    result = []
    remaining = keys
    while remaining:
        batch = remaining[:100]
        remaining = remaining[100:]
        response = get_items_batched(ddb_table_name, batch, client=client, session=session, **kwargs)
        if response.get("Responses") and response["Responses"].get(ddb_table_name):
            result.extend(response["Responses"][ddb_table_name])
        if response.get("UnprocessedKeys") and response["UnprocessedKeys"].get(ddb_table_name):
            remaining = response["UnprocessedKeys"][ddb_table_name]["Keys"] + remaining
    return result


def ddb_query_batch_get_items(
    key_field: str,
    client: Optional[DynamoDBClient] = None,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    **kwargs,
) -> Generator[dict, None, None]:
    table_name = kwargs["TableName"]
    client = client or cast(DynamoDBClient, dynamodb(session=session, config=config).meta.client)
    paginator = client.get_paginator("query")

    page_iterator = paginator.paginate(**kwargs)

    deserializer = TypeDeserializer()

    for page in page_iterator:
        keys = [{key_field: rec[key_field]} for rec in page["Items"]]
        items = ddb_get_items(table_name, keys)
        for item in items:
            yield {k: deserializer.deserialize(v) for k, v in item.items()}


def ddb_query_paginate(
    client: Optional[DynamoDBClient] = None,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    **kwargs,
) -> Generator[dict, None, None]:
    client = client or cast(DynamoDBClient, dynamodb(session=session, config=config).meta.client)
    paginator = client.get_paginator("query")

    page_iterator = paginator.paginate(**kwargs)

    for page in page_iterator:
        yield from page["Items"]


def ddb_query_paginated_count(
    client: Optional[DynamoDBClient] = None,
    session: Optional[Session] = None,
    config: Optional[Config] = None,
    **kwargs,
) -> int:
    client = client or cast(DynamoDBClient, dynamodb(session=session, config=config).meta.client)
    paginator = client.get_paginator("query")

    kwargs = {**kwargs, "Select": "COUNT"}

    page_iterator = paginator.paginate(**kwargs)

    return reduce(lambda acc, page: acc + page["Count"], page_iterator, 0)


def s3_gunzip(
    source_url: str, destination_url: str, session: Optional[Session] = None, config: Optional[Config] = None
):
    """
    Unzip a file on S3 to another S3 location. Streams the file and unzips it with local resources,
    and streams the bytes to the new location. Does not use local disk and minimally uses
    local memory.
    """
    _, src_bucket_name, src_key = s3_split_path(source_url)
    _, dest_bucket_name, dest_key = s3_split_path(destination_url)

    s3obj = s3_object(src_bucket_name, src_key, session=session, config=config)
    down_stream = s3obj.get()["Body"]

    s3obj = s3_bucket(dest_bucket_name, session=session, config=config).put_object(Key=dest_key)
    if not s3obj:
        raise Exception(f"Unknown S3 error when putting empty buffer to {destination_url}") from UnknownServiceError

    with gzip.GzipFile(mode="rb", fileobj=down_stream) as zipfile, S3ObjectWriter(s3obj, encoding=None) as new_obj_buff:
        buf = b" "
        while buf:
            buf = zipfile.read(io.DEFAULT_BUFFER_SIZE * 1000)
            if buf:
                new_obj_buff.write(buf)


def assumed_credentials(
    account_id: str,
    role: str,
    role_session_name: Optional[str] = None,
    duration_seconds: int = 1200,
    sts_endpoint_url: Optional[str] = None,
):
    """
    Refreshes the IAM credentials provided to us by STS for the duration of our session. This
    callback is invoked automatically by boto when we are past the lifetime of our
    session (see DurationSeconds in params).

    Returns:
         dict -> A dictionary containing our new set of credentials from STS as well as the
         expiration timestamp for the session.
    """
    region = os.environ.get("AWS_REGION", "eu-west-2")
    endpoint_url = sts_endpoint_url or f"https://sts.{region}.amazonaws.com"

    role_session_name = role_session_name or f"assumed-{uuid4().hex}"

    sts_client = boto3.client("sts", region_name=region, endpoint_url=endpoint_url)

    params = {
        "RoleArn": f"arn:aws:iam::{account_id}:role/{role}",
        "RoleSessionName": role_session_name,
        "DurationSeconds": duration_seconds,
    }

    response = sts_client.assume_role(**params).get("Credentials")

    credentials = {
        "access_key": response.get("AccessKeyId"),
        "secret_key": response.get("SecretAccessKey"),
        "token": response.get("SessionToken"),
        "expiry_time": response.get("Expiration").isoformat(),
    }
    return credentials


def assumed_role_session(
    account_id: str,
    role: str,
    role_session_name: Optional[str] = None,
    duration_seconds: int = 1200,
    refreshable: bool = False,
    sts_endpoint_url: Optional[str] = None,
) -> Session:
    """
    gets assumed role session for given account id and role name
    :return: Boto session for assumed role
    """

    get_credentials = partial(
        assumed_credentials,
        account_id,
        role,
        role_session_name=role_session_name,
        duration_seconds=duration_seconds,
        sts_endpoint_url=sts_endpoint_url,
    )

    credentials = get_credentials()
    if not refreshable:
        boto_credentials = botocore.credentials.Credentials(
            access_key=credentials["access_key"],
            secret_key=credentials["secret_key"],
            token=credentials["token"],
            method="sts-assume-role",
        )
    else:
        boto_credentials = botocore.credentials.RefreshableCredentials.create_from_metadata(
            metadata=credentials,
            refresh_using=get_credentials,
            method="sts-assume-role",
        )

    session = botocore.session.get_session()

    session._credentials = boto_credentials  # type: ignore[attr-defined]

    return Session(botocore_session=session)


async def async_stream_from_s3(
    s3_get_resp: GetObjectOutputTypeDef, bytes_per_chunk: int
) -> AsyncGenerator[bytes, None]:
    body = s3_get_resp["Body"]
    chunk = await run_in_executor(body.read, bytes_per_chunk)
    while chunk:
        yield chunk
        chunk = await run_in_executor(body.read, bytes_per_chunk)
    if body.closed:
        return

    with contextlib.suppress(Exception):
        body.close()


_RE_CANCELLATION_REASONS = re.compile(r"^.+\[(\w+[^\]]*?)\]$")


def transaction_cancellation_reasons(err: ClientError) -> list[str]:
    """
    get cancellation reasons as strings .. e.g.  ['None', 'ConditionalCheckFailed', 'None']
    """

    if err.response["Error"]["Code"] != _TRANSACTION_CANCELLED:
        return []

    match = _RE_CANCELLATION_REASONS.match(err.response["Error"]["Message"])
    if not match:
        return []

    return [reason.strip() for reason in match.group(1).split(",")]


def transaction_cancelled_for_only_reasons(err: ClientError, *match_reason: str) -> list[str]:
    """
    returns all reasons ... if all reasons either match match_reason or 'None'
    """
    reasons = transaction_cancellation_reasons(err)
    match_reasons = {"None", *list(match_reason)}
    return reasons if all(reason in match_reasons for reason in reasons) else []


def cancellation_reasons_if_conditional_check(err: ClientError) -> list[str]:
    return transaction_cancelled_for_only_reasons(err, "ConditionalCheckFailed")



from collections.abc import Generator

import petname  # type: ignore[import]
import pytest
from mypy_boto3_dynamodb.service_resource import Table

from nhs_aws_helpers import dynamodb
from nhs_aws_helpers.fixtures import temp_dynamodb_table


@pytest.fixture(scope="session", name="source_ddb_table")
def a_ddb_table() -> Generator[Table, None, None]:
    ddb = dynamodb()

    args = {
        "AttributeDefinitions": [
            {"AttributeName": "my_pk", "AttributeType": "S"},
            {"AttributeName": "my_sk", "AttributeType": "S"},
            {"AttributeName": "model_type", "AttributeType": "S"},
        ],
        "KeySchema": [{"AttributeName": "my_pk", "KeyType": "HASH"}, {"AttributeName": "my_sk", "KeyType": "RANGE"}],
        "ProvisionedThroughput": {"ReadCapacityUnits": 500, "WriteCapacityUnits": 500},
        "GlobalSecondaryIndexes": [
            {
                "IndexName": "string",
                "KeySchema": [
                    {"AttributeName": "model_type", "KeyType": "HASH"},
                ],
                "Projection": {
                    "ProjectionType": "KEYS_ONLY",
                },
                "ProvisionedThroughput": {"ReadCapacityUnits": 500, "WriteCapacityUnits": 500},
            },
        ],
    }

    table_name = f"pytest-{petname.Generate(words=4, separator='_')}"

    table = ddb.create_table(TableName=table_name, **args)  # type: ignore[arg-type]

    yield table

    table.delete()


@pytest.fixture(name="cloned_table_defaults")
def _on_demand_table(source_ddb_table: Table) -> Generator[Table, None, None]:
    yield from temp_dynamodb_table(source_ddb_table.table_name)


@pytest.fixture(name="cloned_table_provisioned")
def _provisioned_table(source_ddb_table: Table) -> Generator[Table, None, None]:
    yield from temp_dynamodb_table(source_ddb_table.table_name, on_demand_billing_mode=False, provisioned_capacity=99)


def test_default_temp_table(cloned_table_defaults: Table):
    assert cloned_table_defaults.billing_mode_summary["BillingMode"] == "PAY_PER_REQUEST"


def test_provisioned_temp_table(cloned_table_provisioned: Table):
    assert cloned_table_provisioned.provisioned_throughput == {
        "NumberOfDecreasesToday": 0,
        "ReadCapacityUnits": 99,
        "WriteCapacityUnits": 99,
    }



import sys
from collections.abc import Iterable
from io import SEEK_CUR, SEEK_END, SEEK_SET
from typing import IO, Any, Final, Optional, Union, cast

from botocore.response import StreamingBody
from mypy_boto3_s3.service_resource import Object

from nhs_aws_helpers.common import MiB


class S3ObjectReader(IO):
    """
    provides in-memory buffer to write to S3 objects
    NOTE: ... only use this if you want to buffer the entire stream into memory, to make it seekable
    """

    DEFAULT_BUFFER_SIZE: Final[int] = 8 * MiB

    def __init__(
        self,
        s3_obj: Object,
        encoding: Optional[str] = None,
        line_ending: str = "\n",
        buffer_size: int = DEFAULT_BUFFER_SIZE,
        name: str = "",
    ) -> None:
        self._s3_obj = s3_obj
        self._closed = False
        self._encoding = encoding
        self._buffer: Union[str, bytes] = "" if encoding else b""
        self._bytes_read, self._position = (0, 0)
        self._fully_read = False
        self._body: Optional[StreamingBody] = None
        self._line_ending = line_ending
        self._buffer_size = buffer_size
        self._name = name

    @property
    def s3_obj(self) -> Object:
        return self._s3_obj

    def __iter__(self):
        return self

    def __next__(self):
        line = self.readline()
        if line is None:
            raise StopIteration

        return line.strip()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            return False

        self.close()
        return True

    def write(self, s: Any) -> int:
        raise NotImplementedError

    def writelines(self, lines: Iterable[Any]) -> None:
        raise NotImplementedError

    @property
    def mode(self) -> str:
        return "r" if self._encoding else "rb"

    @property
    def name(self) -> str:
        return self._name

    @property
    def raw(self):
        return self.body

    @property
    def closed(self):
        return self._closed

    @property
    def body(self):
        if self._body is None:
            obj = self._s3_obj.get()
            self._body = obj["Body"]
        return self._body

    def fileno(self):
        return self._s3_obj.key

    def close(self):
        if self._closed:
            return
        if self._body:
            self._body.close()
        self._closed = True

    def flush(self):
        raise NotImplementedError

    def readable(self):
        return self.body and (self._fully_read or not self._closed)

    def readinto(self, byte_array: bytearray) -> int:
        to_read = len(byte_array)
        chunk = self.read(to_read)
        bytes_read = len(chunk)
        for i, char in enumerate(chunk):
            byte_array[i] = char  # type: ignore[call-overload]
        return bytes_read

    def peek(self, num: Optional[int] = None) -> Union[str, bytes]:
        current_pos = self._position
        bytes_read = self.read(num)
        self._position = current_pos
        return bytes_read

    def read(self, num: Optional[int] = None) -> Union[str, bytes]:
        if self._fully_read:
            end = self._bytes_read if num is None else max(min(self._position + num, self._bytes_read), 0)
            pos = min(max(self._position, 0), self._bytes_read)

            chunk = self._buffer[min(pos, end) : max(pos, end)]
            self._position += end - pos
            return chunk

        if num is None:
            chunk = cast(Union[str, bytes], self.body.read())
            if self._encoding:
                chunk = self.ensure_valid_chunk(chunk)
            else:
                self._bytes_read = len(self._buffer) + len(chunk)
            self._buffer = self._buffer + chunk  # type: ignore[operator]
            self._fully_read = True
            self._position = len(self._buffer)
            return chunk

        if num == 0:
            return "" if self._encoding else b""

        end = max(0, self._position + num)

        if end > self._bytes_read:
            bytes_to_read = max(end - self._bytes_read, self._buffer_size)

            chunk = self.body.read(bytes_to_read)
            byte_count = len(chunk)

            if self._encoding:
                chunk = self.ensure_valid_chunk(chunk)
            else:
                self._bytes_read += byte_count

            self._buffer = self._buffer + chunk  # type: ignore[operator]

            # TODO This should really check that you're not at the end of the file
            if byte_count < bytes_to_read:
                self._fully_read = True

        end = min(end, self._bytes_read)

        read_from = min(self._position, end)
        read_to = max(self._position, end)

        self._position = end
        return self._buffer[read_from:read_to]

    def readline(self, limit: int = -1) -> Union[Optional[str], Any]:
        if self._encoding is None:
            raise ValueError("readline only works if _encoding was defined")

        max_read = limit if limit > 0 else sys.maxsize
        chunk_size = min(max_read, 8192)
        bytes_read = 0
        index = -1
        local_buffer = ""
        while index < 0 and bytes_read < max_read:
            chunk = self.read(chunk_size)
            if not chunk:
                if not local_buffer:
                    return None
                break

            local_buffer += chunk  # type: ignore[operator]
            index = local_buffer.find(self._line_ending, bytes_read)
            bytes_read += len(chunk)

        if index < 0:
            return local_buffer

        self.seek((bytes_read - index - 1) * -1, SEEK_CUR)
        return local_buffer[0 : index + 1]

    def readlines(self, size=None):
        if self._encoding is None:
            raise ValueError("readlines only works if _encoding was defined")

        local_buffer = cast(str, self.read(size))

        position = 0
        index = local_buffer.find(self._line_ending, position)
        while index > 0:
            yield local_buffer[position : index + 1]
            next_index = local_buffer.find(self._line_ending, index + 1)
            position = index + 1
            index = next_index

        if position < len(local_buffer):
            yield local_buffer[position:]

    def seek(self, offset: int, whence: int = SEEK_SET) -> int:
        if whence == SEEK_END:
            self.read()  # read full stream
            offset = offset * -1 if offset > 0 else offset
            self._position = max(0, self._bytes_read + offset)
            return self._position

        cur_pos = 0 if whence == SEEK_SET else self._position

        new_pos = max(0, cur_pos + offset)
        if new_pos > self._bytes_read:
            self.read(new_pos - self._position)
            return self._position

        self._position = new_pos
        return self._position

    def seekable(self) -> bool:
        return True

    def isatty(self) -> bool:
        return False

    def tell(self):
        return self._position

    def truncate(self, size=None):
        raise NotImplementedError

    def writable(self) -> bool:
        return False

    def ensure_valid_chunk(self, chunk, run_count: int = 0):
        """We need to check that the chunk is decode-able.
        This is because, with multibyte chars, when we create our chunks, we are just ticking
        through each byte. A multibyte char is between 2 and 4 bytes, so we need to check that
        we have not chopped off a char midway through it's bytes, so will loop through 3 times,
        removing last char and attempting to decode."""

        try:
            chunk = chunk.decode(self._encoding)
            self._bytes_read += len(chunk)
        except UnicodeDecodeError as err:
            if run_count > 2:
                raise err
            run_count += 1
            chunk = chunk[:-1]
            chunk = self.ensure_valid_chunk(chunk, run_count)

        return chunk
